import { afterTick, box, executeCallbacks, mergeProps, onDestroyEffect, useRefById, } from "svelte-toolbelt";
import { Context, IsFocusWithin, watch } from "runed";
import { on } from "svelte/events";
import { FIRST_LAST_KEYS, LAST_KEYS, SELECTION_KEYS, SUB_OPEN_KEYS, getCheckedState, isMouseEvent, } from "./utils.js";
import { focusFirst } from "../../internal/focus.js";
import { CustomEventDispatcher } from "../../internal/events.js";
import { useDOMTypeahead } from "../../internal/use-dom-typeahead.svelte.js";
import { isElement, isElementOrSVGElement, isHTMLElement } from "../../internal/is.js";
import { useRovingFocus } from "../../internal/use-roving-focus.svelte.js";
import { kbd } from "../../internal/kbd.js";
import { getAriaChecked, getAriaDisabled, getAriaExpanded, getAriaOrientation, getDataDisabled, getDataOpenClosed, } from "../../internal/attrs.js";
import { isPointerInGraceArea, makeHullFromElements } from "../../internal/polygon.js";
export const CONTEXT_MENU_TRIGGER_ATTR = "data-context-menu-trigger";
const MenuRootContext = new Context("Menu.Root");
const MenuMenuContext = new Context("Menu.Root | Menu.Sub");
const MenuContentContext = new Context("Menu.Content");
const MenuGroupContext = new Context("Menu.Group | Menu.RadioGroup");
const MenuRadioGroupContext = new Context("Menu.RadioGroup");
export const MenuOpenEvent = new CustomEventDispatcher("bitsmenuopen", {
    bubbles: false,
    cancelable: true,
});
class MenuRootState {
    onClose;
    variant;
    isUsingKeyboard = $state(false);
    dir;
    constructor(props) {
        this.onClose = props.onClose;
        this.dir = props.dir;
        this.variant = props.variant;
        $effect(() => {
            const callbacksToDispose = [];
            const handlePointer = (_) => {
                this.isUsingKeyboard = false;
            };
            const handleKeydown = (_) => {
                this.isUsingKeyboard = true;
                const disposePointerDown = on(document, "pointerdown", handlePointer, {
                    capture: true,
                    once: true,
                });
                const disposePointerMove = on(document, "pointermove", handlePointer, {
                    capture: true,
                    once: true,
                });
                callbacksToDispose.push(disposePointerDown, disposePointerMove);
            };
            const disposeKeydown = on(document, "keydown", handleKeydown, {
                capture: true,
            });
            callbacksToDispose.push(disposeKeydown);
            return () => {
                executeCallbacks(callbacksToDispose);
            };
        });
    }
    getAttr(name) {
        return `data-${this.variant.current}-${name}`;
    }
}
class MenuMenuState {
    root;
    open;
    contentId = box.with(() => "");
    contentNode = $state(null);
    triggerNode = $state(null);
    parentMenu;
    constructor(props, root, parentMenu) {
        this.root = root;
        this.open = props.open;
        this.parentMenu = parentMenu;
        if (parentMenu) {
            watch(() => parentMenu.open, (isOpen) => {
                if (isOpen)
                    return;
                this.open.current = false;
            });
        }
    }
    toggleOpen() {
        this.open.current = !this.open.current;
    }
    onOpen() {
        this.open.current = true;
    }
    onClose() {
        this.open.current = false;
    }
}
class MenuContentState {
    #id;
    contentRef;
    parentMenu;
    search = $state("");
    #loop;
    #timer = $state(0);
    pointerGraceTimer = $state(0);
    #pointerGraceIntent = $state(null);
    #pointerDir = $state("right");
    #lastPointerX = $state(0);
    #handleTypeaheadSearch;
    rovingFocusGroup;
    isMounted;
    isFocusWithin = new IsFocusWithin(() => this.parentMenu.contentNode ?? undefined);
    constructor(props, parentMenu) {
        this.#id = props.id;
        this.#loop = props.loop;
        this.parentMenu = parentMenu;
        this.parentMenu.contentId = props.id;
        this.contentRef = props.ref;
        this.isMounted = props.isMounted;
        this.onkeydown = this.onkeydown.bind(this);
        this.onblur = this.onblur.bind(this);
        this.onpointermove = this.onpointermove.bind(this);
        this.onfocus = this.onfocus.bind(this);
        this.handleInteractOutside = this.handleInteractOutside.bind(this);
        useRefById({
            id: this.#id,
            ref: this.contentRef,
            deps: () => this.parentMenu.open.current,
            onRefChange: (node) => {
                if (this.parentMenu.contentNode !== node) {
                    this.parentMenu.contentNode = node;
                }
            },
        });
        onDestroyEffect(() => {
            window.clearTimeout(this.#timer);
        });
        this.#handleTypeaheadSearch = useDOMTypeahead().handleTypeaheadSearch;
        this.rovingFocusGroup = useRovingFocus({
            rootNodeId: this.parentMenu.contentId,
            candidateAttr: this.parentMenu.root.getAttr("item"),
            loop: this.#loop,
            orientation: box.with(() => "vertical"),
        });
        watch(() => this.parentMenu.contentNode, (contentNode) => {
            if (!contentNode)
                return;
            const handler = () => {
                afterTick(() => {
                    if (!this.parentMenu.root.isUsingKeyboard)
                        return;
                    this.rovingFocusGroup.focusFirstCandidate();
                });
            };
            return MenuOpenEvent.listen(contentNode, handler);
        });
    }
    #getCandidateNodes() {
        const node = this.parentMenu.contentNode;
        if (!node)
            return [];
        const candidates = Array.from(node.querySelectorAll(`[${this.parentMenu.root.getAttr("item")}]:not([data-disabled])`));
        return candidates;
    }
    #isPointerMovingToSubmenu(e) {
        const isMovingTowards = this.#pointerDir === this.#pointerGraceIntent?.side;
        return isMovingTowards && isPointerInGraceArea(e, this.#pointerGraceIntent?.area);
    }
    onPointerGraceIntentChange(intent) {
        this.#pointerGraceIntent = intent;
    }
    onkeydown(e) {
        if (e.defaultPrevented)
            return;
        const target = e.target;
        const currentTarget = e.currentTarget;
        if (!isHTMLElement(target) || !isHTMLElement(currentTarget))
            return;
        const isKeydownInside = target.closest(`[${this.parentMenu.root.getAttr("content")}]`)?.id ===
            this.parentMenu.contentId.current;
        const isModifierKey = e.ctrlKey || e.altKey || e.metaKey;
        const isCharacterKey = e.key.length === 1;
        const kbdFocusedEl = this.rovingFocusGroup.handleKeydown(target, e);
        if (kbdFocusedEl)
            return;
        // prevent space from being considered with typeahead
        if (e.code === "Space")
            return;
        const candidateNodes = this.#getCandidateNodes();
        if (isKeydownInside) {
            // menus do not respect the tab key
            if (e.key === kbd.TAB)
                e.preventDefault();
            if (!isModifierKey && isCharacterKey) {
                this.#handleTypeaheadSearch(e.key, candidateNodes);
            }
        }
        // focus first/last based on key pressed
        if (e.target?.id !== this.parentMenu.contentId.current)
            return;
        if (!FIRST_LAST_KEYS.includes(e.key))
            return;
        e.preventDefault();
        if (LAST_KEYS.includes(e.key)) {
            candidateNodes.reverse();
        }
        focusFirst(candidateNodes);
    }
    onblur(e) {
        if (!isElement(e.currentTarget))
            return;
        if (!isElement(e.target))
            return;
        // clear search buffer when leaving the menu
        if (!e.currentTarget.contains?.(e.target)) {
            window.clearTimeout(this.#timer);
            this.search = "";
        }
    }
    onfocus(_) {
        if (!this.parentMenu.root.isUsingKeyboard)
            return;
        afterTick(() => this.rovingFocusGroup.focusFirstCandidate());
    }
    onpointermove(e) {
        if (!isMouseEvent(e))
            return;
        const target = e.target;
        if (!isElement(target))
            return;
        const pointerXHasChanged = this.#lastPointerX !== e.clientX;
        const currentTarget = e.currentTarget;
        if (!isElement(currentTarget))
            return;
        // We don't use `event.movementX` for this check because Safari will
        // always return `0` on a pointer event.
        if (currentTarget.contains(target) && pointerXHasChanged) {
            const newDir = e.clientX > this.#lastPointerX ? "right" : "left";
            this.#pointerDir = newDir;
            this.#lastPointerX = e.clientX;
        }
    }
    onItemEnter(e) {
        if (this.#isPointerMovingToSubmenu(e))
            return true;
        return false;
    }
    onItemLeave(e) {
        if (this.#isPointerMovingToSubmenu(e))
            return;
        const contentNode = this.parentMenu.contentNode;
        contentNode?.focus();
        this.rovingFocusGroup.setCurrentTabStopId("");
    }
    onTriggerLeave(e) {
        if (this.#isPointerMovingToSubmenu(e))
            return true;
        return false;
    }
    onOpenAutoFocus = (e) => {
        if (e.defaultPrevented)
            return;
        e.preventDefault();
        const contentNode = this.parentMenu.contentNode;
        contentNode?.focus();
    };
    handleInteractOutside(e) {
        if (!isElementOrSVGElement(e.target))
            return;
        const triggerId = this.parentMenu.triggerNode?.id;
        if (e.target.id === triggerId) {
            e.preventDefault();
            return;
        }
        if (e.target.closest(`#${triggerId}`)) {
            e.preventDefault();
        }
    }
    snippetProps = $derived.by(() => ({ open: this.parentMenu.open.current }));
    props = $derived.by(() => ({
        id: this.#id.current,
        role: "menu",
        "aria-orientation": getAriaOrientation("vertical"),
        [this.parentMenu.root.getAttr("content")]: "",
        "data-state": getDataOpenClosed(this.parentMenu.open.current),
        onkeydown: this.onkeydown,
        onblur: this.onblur,
        onpointermove: this.onpointermove,
        onfocus: this.onfocus,
        dir: this.parentMenu.root.dir.current,
        style: {
            pointerEvents: "auto",
        },
    }));
}
class MenuItemSharedState {
    content;
    ref;
    id;
    disabled;
    #isFocused = $state(false);
    constructor(props, content) {
        this.content = content;
        this.id = props.id;
        this.disabled = props.disabled;
        this.ref = props.ref;
        this.onpointermove = this.onpointermove.bind(this);
        this.onpointerleave = this.onpointerleave.bind(this);
        this.onfocus = this.onfocus.bind(this);
        this.onblur = this.onblur.bind(this);
        useRefById({
            id: this.id,
            ref: this.ref,
            deps: () => this.content.isMounted.current,
        });
    }
    onpointermove(e) {
        if (e.defaultPrevented)
            return;
        if (!isMouseEvent(e))
            return;
        if (this.disabled.current) {
            this.content.onItemLeave(e);
        }
        else {
            const defaultPrevented = this.content.onItemEnter(e);
            if (defaultPrevented)
                return;
            const item = e.currentTarget;
            if (!isHTMLElement(item))
                return;
            item.focus();
        }
    }
    onpointerleave(e) {
        afterTick(() => {
            if (e.defaultPrevented)
                return;
            if (!isMouseEvent(e))
                return;
            this.content.onItemLeave(e);
        });
    }
    onfocus(e) {
        afterTick(() => {
            if (e.defaultPrevented || this.disabled.current)
                return;
            this.#isFocused = true;
        });
    }
    onblur(e) {
        afterTick(() => {
            if (e.defaultPrevented)
                return;
            this.#isFocused = false;
        });
    }
    props = $derived.by(() => ({
        id: this.id.current,
        tabindex: -1,
        role: "menuitem",
        "aria-disabled": getAriaDisabled(this.disabled.current),
        "data-disabled": getDataDisabled(this.disabled.current),
        "data-highlighted": this.#isFocused ? "" : undefined,
        [this.content.parentMenu.root.getAttr("item")]: "",
        //
        onpointermove: this.onpointermove,
        onpointerleave: this.onpointerleave,
        onfocus: this.onfocus,
        onblur: this.onblur,
    }));
}
class MenuItemState {
    #item;
    #onSelect;
    #closeOnSelect;
    #isPointerDown = $state(false);
    root;
    constructor(props, item) {
        this.#item = item;
        this.root = item.content.parentMenu.root;
        this.#onSelect = props.onSelect;
        this.#closeOnSelect = props.closeOnSelect;
        this.onkeydown = this.onkeydown.bind(this);
        this.onclick = this.onclick.bind(this);
        this.onpointerdown = this.onpointerdown.bind(this);
        this.onpointerup = this.onpointerup.bind(this);
    }
    #handleSelect() {
        if (this.#item.disabled.current)
            return;
        const selectEvent = new CustomEvent("menuitemselect", { bubbles: true, cancelable: true });
        this.#onSelect.current(selectEvent);
        afterTick(() => {
            if (selectEvent.defaultPrevented) {
                this.#item.content.parentMenu.root.isUsingKeyboard = false;
                return;
            }
            if (this.#closeOnSelect.current) {
                this.#item.content.parentMenu.root.onClose();
            }
        });
    }
    onkeydown(e) {
        const isTypingAhead = this.#item.content.search !== "";
        if (this.#item.disabled.current || (isTypingAhead && e.key === kbd.SPACE))
            return;
        if (SELECTION_KEYS.includes(e.key)) {
            if (!isHTMLElement(e.currentTarget))
                return;
            e.currentTarget.click();
            /**
             * We prevent default browser behavior for selection keys as they should trigger
             * a selection only:
             * - prevents space from scrolling the page.
             * - if keydown causes focus to move, prevents keydown from firing on the new target.
             */
            e.preventDefault();
        }
    }
    onclick(_) {
        if (this.#item.disabled.current)
            return;
        this.#handleSelect();
    }
    onpointerup(e) {
        if (e.defaultPrevented)
            return;
        if (!this.#isPointerDown) {
            if (!isHTMLElement(e.currentTarget))
                return;
            e.currentTarget?.click();
        }
    }
    onpointerdown(_) {
        this.#isPointerDown = true;
    }
    props = $derived.by(() => mergeProps(this.#item.props, {
        onclick: this.onclick,
        onpointerdown: this.onpointerdown,
        onpointerup: this.onpointerup,
        onkeydown: this.onkeydown,
    }));
}
class MenuSubTriggerState {
    #item;
    // The menu this sub-trigger item belongs within
    #content;
    // the menu this sub-trigger item opens
    #submenu;
    #openTimer = $state(null);
    constructor(item, content, submenu) {
        this.#item = item;
        this.#content = content;
        this.#submenu = submenu;
        this.onpointerleave = this.onpointerleave.bind(this);
        this.onpointermove = this.onpointermove.bind(this);
        this.onkeydown = this.onkeydown.bind(this);
        this.onclick = this.onclick.bind(this);
        onDestroyEffect(() => {
            this.#clearOpenTimer();
        });
        useRefById({
            id: this.#item.id,
            ref: this.#item.ref,
            onRefChange: (node) => {
                this.#submenu.triggerNode = node;
            },
        });
    }
    #clearOpenTimer() {
        if (this.#openTimer === null)
            return;
        window.clearTimeout(this.#openTimer);
        this.#openTimer = null;
    }
    onpointermove(e) {
        if (!isMouseEvent(e))
            return;
        const defaultPrevented = this.#content.onItemEnter(e);
        if (defaultPrevented)
            return;
        if (!this.#item.disabled.current && !this.#submenu.open.current && !this.#openTimer) {
            this.#content.onPointerGraceIntentChange(null);
            this.#openTimer = window.setTimeout(() => {
                this.#submenu.onOpen();
                this.#clearOpenTimer();
            }, 100);
        }
    }
    onpointerleave(e) {
        if (!isMouseEvent(e))
            return;
        this.#clearOpenTimer();
        const contentNode = this.#submenu.contentNode;
        const subTriggerNode = this.#item.ref.current;
        if (contentNode && subTriggerNode) {
            const polygon = makeHullFromElements([subTriggerNode, contentNode]);
            const side = contentNode?.dataset.side;
            this.#content.onPointerGraceIntentChange({
                area: polygon,
                side,
            });
            window.clearTimeout(this.#content.pointerGraceTimer);
            this.#content.pointerGraceTimer = window.setTimeout(() => this.#content.onPointerGraceIntentChange(null), 300);
        }
        else {
            const defaultPrevented = this.#content.onTriggerLeave(e);
            if (defaultPrevented)
                return;
            // There's 100ms where the user may leave an item before the submenu was opened.
            this.#content.onPointerGraceIntentChange(null);
        }
    }
    onkeydown(e) {
        const isTypingAhead = this.#content.search !== "";
        if (this.#item.disabled.current || (isTypingAhead && e.key === kbd.SPACE))
            return;
        if (SUB_OPEN_KEYS[this.#submenu.root.dir.current].includes(e.key)) {
            this.#submenu.onOpen();
            afterTick(() => {
                const contentNode = this.#submenu.contentNode;
                if (!contentNode)
                    return;
                MenuOpenEvent.dispatch(contentNode);
            });
            e.preventDefault();
        }
    }
    onclick(e) {
        if (this.#item.disabled.current)
            return;
        /**
         * We manually focus because iOS Safari doesn't always focus on click (e.g. buttons)
         * and we rely heavily on `onFocusOutside` for submenus to close when switching
         * between separate submenus.
         */
        if (!isHTMLElement(e.currentTarget))
            return;
        e.currentTarget.focus();
        if (!this.#submenu.open.current) {
            this.#submenu.onOpen();
        }
    }
    props = $derived.by(() => mergeProps({
        "aria-haspopup": "menu",
        "aria-expanded": getAriaExpanded(this.#submenu.open.current),
        "data-state": getDataOpenClosed(this.#submenu.open.current),
        "aria-controls": this.#submenu.open.current
            ? this.#submenu.contentId.current
            : undefined,
        [this.#submenu.root.getAttr("sub-trigger")]: "",
        onclick: this.onclick,
        onpointermove: this.onpointermove,
        onpointerleave: this.onpointerleave,
        onkeydown: this.onkeydown,
    }, this.#item.props));
}
class MenuCheckboxItemState {
    #item;
    #checked;
    #indeterminate;
    constructor(props, item) {
        this.#item = item;
        this.#checked = props.checked;
        this.#indeterminate = props.indeterminate;
    }
    toggleChecked() {
        if (this.#indeterminate.current) {
            this.#indeterminate.current = false;
            this.#checked.current = true;
        }
        else {
            this.#checked.current = !this.#checked.current;
        }
    }
    snippetProps = $derived.by(() => ({
        checked: this.#checked.current,
        indeterminate: this.#indeterminate.current,
    }));
    props = $derived.by(() => ({
        ...this.#item.props,
        role: "menuitemcheckbox",
        "aria-checked": getAriaChecked(this.#checked.current, this.#indeterminate.current),
        "data-state": getCheckedState(this.#checked.current),
        [this.#item.root.getAttr("checkbox-item")]: "",
    }));
}
class MenuGroupState {
    #id;
    #ref;
    groupHeadingId = $state(undefined);
    root;
    constructor(props, root) {
        this.#id = props.id;
        this.#ref = props.ref;
        this.root = root;
        useRefById({
            id: this.#id,
            ref: this.#ref,
        });
    }
    props = $derived.by(() => ({
        id: this.#id.current,
        role: "group",
        "aria-labelledby": this.groupHeadingId,
        [this.root.getAttr("group")]: "",
    }));
}
class MenuGroupHeadingState {
    #id;
    #ref;
    #group;
    constructor(props, group) {
        this.#id = props.id;
        this.#ref = props.ref;
        this.#group = group;
        useRefById({
            id: this.#id,
            ref: this.#ref,
            onRefChange: (node) => {
                this.#group.groupHeadingId = node?.id;
            },
        });
    }
    props = $derived.by(() => ({
        id: this.#id.current,
        role: "group",
        [this.#group.root.getAttr("group-heading")]: "",
    }));
}
class MenuSeparatorState {
    #id;
    #ref;
    #root;
    constructor(props, root) {
        this.#id = props.id;
        this.#ref = props.ref;
        this.#root = root;
        useRefById({
            id: this.#id,
            ref: this.#ref,
        });
    }
    props = $derived.by(() => ({
        id: this.#id.current,
        role: "group",
        [this.#root.getAttr("separator")]: "",
    }));
}
class MenuArrowState {
    #root;
    constructor(root) {
        this.#root = root;
    }
    props = $derived.by(() => ({
        [this.#root.getAttr("arrow")]: "",
    }));
}
class MenuRadioGroupState {
    #id;
    value;
    #ref;
    content;
    groupHeadingId = $state(null);
    root;
    constructor(props, content) {
        this.value = props.value;
        this.#id = props.id;
        this.#ref = props.ref;
        this.content = content;
        this.root = content.parentMenu.root;
        useRefById({
            id: this.#id,
            ref: this.#ref,
        });
    }
    setValue(v) {
        this.value.current = v;
    }
    props = $derived.by(() => ({
        id: this.#id.current,
        [this.root.getAttr("radio-group")]: "",
        role: "group",
        "aria-labelledby": this.groupHeadingId,
    }));
}
class MenuRadioItemState {
    #id;
    #ref;
    #item;
    #value;
    #group;
    isChecked = $derived.by(() => this.#group.value.current === this.#value.current);
    constructor(props, item, group) {
        this.#item = item;
        this.#id = props.id;
        this.#ref = props.ref;
        this.#group = group;
        this.#value = props.value;
        useRefById({
            id: this.#id,
            ref: this.#ref,
        });
    }
    selectValue() {
        this.#group.setValue(this.#value.current);
    }
    props = $derived.by(() => ({
        [this.#group.root.getAttr("radio-item")]: "",
        ...this.#item.props,
        role: "menuitemradio",
        "aria-checked": getAriaChecked(this.isChecked, false),
        "data-state": getCheckedState(this.isChecked),
    }));
}
class DropdownMenuTriggerState {
    #id;
    #ref;
    #parentMenu;
    #disabled;
    constructor(props, parentMenu) {
        this.#ref = props.ref;
        this.#id = props.id;
        this.#parentMenu = parentMenu;
        this.#disabled = props.disabled;
        this.onpointerdown = this.onpointerdown.bind(this);
        this.onpointerup = this.onpointerup.bind(this);
        this.onkeydown = this.onkeydown.bind(this);
        useRefById({
            id: this.#id,
            ref: this.#ref,
            onRefChange: (ref) => {
                this.#parentMenu.triggerNode = ref;
            },
        });
    }
    onpointerdown(e) {
        if (this.#disabled.current)
            return;
        if (e.pointerType === "touch")
            return e.preventDefault();
        if (e.button === 0 && e.ctrlKey === false) {
            this.#parentMenu.toggleOpen();
            // prevent trigger focusing when opening to allow
            // the content to be given focus without competition
            if (!this.#parentMenu.open.current)
                e.preventDefault();
        }
    }
    onpointerup(e) {
        if (this.#disabled.current)
            return;
        if (e.pointerType === "touch") {
            e.preventDefault();
            this.#parentMenu.toggleOpen();
        }
    }
    onkeydown(e) {
        if (this.#disabled.current)
            return;
        if (e.key === kbd.SPACE || e.key === kbd.ENTER) {
            this.#parentMenu.toggleOpen();
            e.preventDefault();
            return;
        }
        if (e.key === kbd.ARROW_DOWN) {
            this.#parentMenu.onOpen();
            e.preventDefault();
        }
    }
    #ariaControls = $derived.by(() => {
        if (this.#parentMenu.open.current && this.#parentMenu.contentId.current)
            return this.#parentMenu.contentId.current;
        return undefined;
    });
    props = $derived.by(() => ({
        id: this.#id.current,
        disabled: this.#disabled.current,
        "aria-haspopup": "menu",
        "aria-expanded": getAriaExpanded(this.#parentMenu.open.current),
        "aria-controls": this.#ariaControls,
        "data-disabled": getDataDisabled(this.#disabled.current),
        "data-state": getDataOpenClosed(this.#parentMenu.open.current),
        [this.#parentMenu.root.getAttr("trigger")]: "",
        //
        onpointerdown: this.onpointerdown,
        onpointerup: this.onpointerup,
        onkeydown: this.onkeydown,
    }));
}
class ContextMenuTriggerState {
    #id;
    #ref;
    #parentMenu;
    #disabled;
    #point = $state({ x: 0, y: 0 });
    virtualElement = box({
        getBoundingClientRect: () => DOMRect.fromRect({ width: 0, height: 0, ...this.#point }),
    });
    #longPressTimer = $state(null);
    constructor(props, parentMenu) {
        this.#parentMenu = parentMenu;
        this.#disabled = props.disabled;
        this.#id = props.id;
        this.#ref = props.ref;
        this.oncontextmenu = this.oncontextmenu.bind(this);
        this.onpointerdown = this.onpointerdown.bind(this);
        this.onpointermove = this.onpointermove.bind(this);
        this.onpointercancel = this.onpointercancel.bind(this);
        this.onpointerup = this.onpointerup.bind(this);
        useRefById({
            id: this.#id,
            ref: this.#ref,
            onRefChange: (node) => {
                this.#parentMenu.triggerNode = node;
            },
            deps: () => this.#parentMenu.open.current,
        });
        watch(() => this.#point, (point) => {
            this.virtualElement.current = {
                getBoundingClientRect: () => DOMRect.fromRect({ width: 0, height: 0, ...point }),
            };
        });
        $effect(() => {
            if (this.#disabled.current) {
                this.#clearLongPressTimer();
            }
        });
        onDestroyEffect(() => this.#clearLongPressTimer());
    }
    #clearLongPressTimer() {
        if (this.#longPressTimer === null)
            return;
        window.clearTimeout(this.#longPressTimer);
    }
    #handleOpen(e) {
        this.#point = { x: e.clientX, y: e.clientY };
        this.#parentMenu.onOpen();
    }
    oncontextmenu(e) {
        if (this.#disabled.current)
            return;
        this.#clearLongPressTimer();
        this.#handleOpen(e);
        e.preventDefault();
        this.#parentMenu.contentNode?.focus();
    }
    onpointerdown(e) {
        if (this.#disabled.current || isMouseEvent(e))
            return;
        this.#clearLongPressTimer();
        this.#longPressTimer = window.setTimeout(() => this.#handleOpen(e), 700);
    }
    onpointermove(e) {
        if (this.#disabled.current || isMouseEvent(e))
            return;
        this.#clearLongPressTimer();
    }
    onpointercancel(e) {
        if (this.#disabled.current || isMouseEvent(e))
            return;
        this.#clearLongPressTimer();
    }
    onpointerup(e) {
        if (this.#disabled.current || isMouseEvent(e))
            return;
        this.#clearLongPressTimer();
    }
    props = $derived.by(() => ({
        id: this.#id.current,
        disabled: this.#disabled.current,
        "data-disabled": getDataDisabled(this.#disabled.current),
        "data-state": getDataOpenClosed(this.#parentMenu.open.current),
        [CONTEXT_MENU_TRIGGER_ATTR]: "",
        //
        onpointerdown: this.onpointerdown,
        onpointermove: this.onpointermove,
        onpointercancel: this.onpointercancel,
        onpointerup: this.onpointerup,
        oncontextmenu: this.oncontextmenu,
    }));
}
export function useMenuRoot(props) {
    return MenuRootContext.set(new MenuRootState(props));
}
export function useMenuMenu(root, props) {
    return MenuMenuContext.set(new MenuMenuState(props, root));
}
export function useMenuSubmenu(props) {
    const menu = MenuMenuContext.get();
    return MenuMenuContext.set(new MenuMenuState(props, menu.root, menu));
}
export function useMenuSubTrigger(props) {
    const content = MenuContentContext.get();
    const item = new MenuItemSharedState(props, content);
    const submenu = MenuMenuContext.get();
    return new MenuSubTriggerState(item, content, submenu);
}
export function useMenuDropdownTrigger(props) {
    return new DropdownMenuTriggerState(props, MenuMenuContext.get());
}
export function useMenuContextTrigger(props) {
    return new ContextMenuTriggerState(props, MenuMenuContext.get());
}
export function useMenuContent(props) {
    return MenuContentContext.set(new MenuContentState(props, MenuMenuContext.get()));
}
export function useMenuItem(props) {
    const item = new MenuItemSharedState(props, MenuContentContext.get());
    return new MenuItemState(props, item);
}
export function useMenuCheckboxItem(props) {
    const item = new MenuItemState(props, new MenuItemSharedState(props, MenuContentContext.get()));
    return new MenuCheckboxItemState(props, item);
}
export function useMenuRadioGroup(props) {
    return MenuGroupContext.set(MenuRadioGroupContext.set(new MenuRadioGroupState(props, MenuContentContext.get())));
}
export function useMenuRadioItem(props) {
    const radioGroup = MenuRadioGroupContext.get();
    const sharedItem = new MenuItemSharedState(props, radioGroup.content);
    const item = new MenuItemState(props, sharedItem);
    return new MenuRadioItemState(props, item, radioGroup);
}
export function useMenuGroup(props) {
    return MenuGroupContext.set(new MenuGroupState(props, MenuRootContext.get()));
}
export function useMenuGroupHeading(props) {
    return new MenuGroupHeadingState(props, MenuGroupContext.get());
}
export function useMenuSeparator(props) {
    return new MenuSeparatorState(props, MenuRootContext.get());
}
export function useMenuArrow() {
    return new MenuArrowState(MenuRootContext.get());
}
